/******************************************************************************/
/*                                                                            */
/* Proyecto final de carrera: AIIV                                            */
/* rea:                      Inteligencia Artificial                         */
/* Universiadad:              UOC                                             */
/* Autor:                     Pedro Amador Daz                               */
/* Fecha:                     24/10/2011                                      */
/*                                                                            */
/******************************************************************************/

#include <16F876.h>

#device ADC=8 /* Resolucin del modulo A/D de 8 bits */
#fuses XT,NOWDT,PUT,NOPROTECT,NOLVP,BROWNOUT /* Configuracin microcontrolador */
#use delay(clock=4000000) /* Frecuencia del oscilador (cristal de cuarzo) */
#use rs232(baud=19200, xmit=pin_c6, rcv=pin_c7) /* Configuracin RS-232 */
#byte puerto_a = 05 /* Direccin del puerto A */
#byte puerto_b = 06 /* Direccin del puerto B */
#byte puerto_c = 07 /* Direccin del puerto c */

#include <stdlib.h>

/* Variables globales */
int maxMinScanner[3][2]= {{7,230},{7,222},{9,235}}; /* Valores mximos y mnimos */
int arrayScanner[3]= {0, 0, 0}; /* Valores ledos por el scanner */
unsigned char selected[5][5]= {{0,255}, {1,255}, {2,255}, {3,255}, {4,255}}; /* Cromosomas seleccionados (indice y calidad) */
unsigned char PWM= 50; /* Potencia mxima aplicable a los actuadores 0-127 */
unsigned char populationIndex= 0; /* Indice de los individuos de la poblacin */
unsigned char chromosome[3]= {100, 100, 100}; /* Combinacin de genes inicial */
unsigned char populationLimit = 0; /* Puntero al ltimo individuo de la poblacin  */
unsigned char life= 0; /* Vida del individuo actual */
signed long error= 0, quality= 0;
float Kp= 0.3, Ki= 0.3, Kd= 0.01;
enum {REPRODUCTION, EVALUATION, SELECTION} state;


/******************************************************************************/
/*                                                                            */
/* motor1                                                                     */
/* Descripcin:      Control de la potencia y sentido de giro del motor 1     */
/* Entrada:          De -127 a 127                                            */
/* Salida:                                                                    */
/* Auto:             Pedro Amador Daz                                        */
/* Fecha:            24/10/2011                                               */
/*                                                                            */
/******************************************************************************/

void motor1 (signed long power)
{
   /* Tratamiento del desbordamiento del parametro de entrada */
   if (power>=127)
      power = 127;
   else if (power<=-127)
      power = -127;
   
   /* Control de la potencia mediante PWM */
   set_pwm1_duty(abs((signed int)power)*2);
   
   if (power>=0)
   { 
      bit_clear(puerto_b,4);
      bit_set(puerto_b,5); 
   }
   else /* Para valores negativos, inversin de giro del motor */
   { 
      bit_set(puerto_b,4);
      bit_clear(puerto_b,5); 
   }
}


/******************************************************************************/
/*                                                                            */
/* motor2                                                                     */
/* Descripcin:      Control de la potencia y sentido de giro del motor 2     */
/* Entrada:          De -127 a 127                                            */
/* Salida:                                                                    */
/* Auto:             Pedro Amador Daz                                        */
/* Fecha:            24/10/2011                                               */
/*                                                                            */
/******************************************************************************/

void motor2 (signed long power)
{
   /* Tratamiento del desbordamiento del parametro de entrada */
   if (power>=127)
      power = 127;
   else if (power<=-127)
      power = -127;
      
   set_pwm2_duty(abs((signed int)power)*2);   /* Control de la potencia mediante PWM */
   
   if (power>=0)
   { 
      bit_clear(puerto_b,6);
      bit_set(puerto_b,7); 
   }
   else /* Para valores negativos, inversin de giro del motor */
   { 
      bit_set(puerto_b,6);
      bit_clear(puerto_b,7); 
   }
}

/******************************************************************************/
/*                                                                            */
/* mutation                                                                   */
/* Descripcin:      Muta los genes del cromosoma                             */
/* Entrada:                                                                   */
/* Salida:                                                                    */
/* Auto:             Pedro Amador Daz                                        */
/* Fecha:            01/01/2012                                               */
/*                                                                            */
/******************************************************************************/

void mutation()
{
   switch (rand()/10923)
   {
      case 0:
         chromosome[0]= rand()/164;
         break;
      case 1:
         chromosome[1]= rand()/164;
         break;
      case 2:
         chromosome[2]= rand()/164;
         break;   
   }
}

/******************************************************************************/
/*                                                                            */
/* chromosomeReproduction                                                     */
/* Descripcin:      Reproduce un nuevo cromosoma a partir de dos padres      */
/*                   e introduce una mutcin en uno de sus genes              */
/* Entrada:          Puntero a los cromosomas padres y posicin donde se      */
/*                   almacenar el cromosoma del nuevo individuo              */
/* Salida:                                                                    */
/* Auto:             Pedro Amador Daz                                        */
/* Fecha:            01/01/2012                                               */
/*                                                                            */
/******************************************************************************/

void chromosomeReproduction(unsigned char parent1, unsigned char parent2, unsigned char reproducedChromosome)
{
         chromosome[0]= (unsigned char)(((unsigned long)read_eeprom((parent1*4)+0)+(unsigned long)read_eeprom((parent2*4)+0))/(unsigned long)2);
         chromosome[1]= (unsigned char)(((unsigned long)read_eeprom((parent1*4)+1)+(unsigned long)read_eeprom((parent2*4)+1))/(unsigned long)2);
         chromosome[2]= (unsigned char)(((unsigned long)read_eeprom((parent1*4)+2)+(unsigned long)read_eeprom((parent2*4)+2))/(unsigned long)2);
         
         mutation();
         
         write_eeprom(((reproducedChromosome)*4)+0,chromosome[0]);
         write_eeprom(((reproducedChromosome)*4)+1,chromosome[1]);
         write_eeprom(((reproducedChromosome)*4)+2,chromosome[2]);
         write_eeprom(((reproducedChromosome)*4)+3,0);
}

/******************************************************************************/
/*                                                                            */
/* geneticAlgorithm                                                           */
/* Descripcin:      Mquina de estados que forma el algoritmo gentico       */
/* Entrada:                                                                   */
/* Salida:                                                                    */
/* Auto:             Pedro Amador Daz                                        */
/* Fecha:            01/01/2012                                               */
/*                                                                            */
/******************************************************************************/

void geneticAlgorithm()
{   
   unsigned char index1=0, index2= 0;
   struct {
      unsigned char index;
      unsigned char value;
   } minQuality;
   
   switch (state)
   {
      case EVALUATION:
         
      /* Calculo del valor calidad, con el que se evaluar el exito del individuo */
      quality = quality + abs(error);
   
      /* Termin la vida del actual individuo */
      if (life>=250)
      { 
         /* Guardamos el valor de calidad del individuo evaluado */
         write_eeprom((populationIndex*4)+3,(int)(quality/250));
         
      /* Siguiente cromosoma para un nuevo individuo */ 
      populationIndex++;
      life = 0;
      quality = 0;
      }
      
      if (populationIndex==populationLimit)
      {
         state= SELECTION;
         break;
      }
      
      /* Ha nacido un nuevo individuo */
      if (life==0)
      { 
         /* A partir de los genes formamos el fenotipo del nuevo individuo */
         Kp= ((unsigned long)read_eeprom((populationIndex*4)+0)+200)/1000.0;
         Ki= ((unsigned long)read_eeprom((populationIndex*4)+1)+200)/1000.0;
         Kd= (read_eeprom((populationIndex*4)+2))/10000.0;
         printf ("Kp= %1.3f, Ki= %1.3f, Kd= %1.3f, Quality= %d\n", Kp, Ki, Kd, (int)(quality/250));
      }
      
      /* El individuo en evaluacin envejece */
      ++life;
      
         break;
         
      case SELECTION:
            
            /* Dos bucles de busqueda anidados para actualizar el array de los mejores 5 individuos de la poblacin */
            index1= populationIndex-10;
            for (index1; index1<populationLimit; index1++)
            {
               /* Buscamos el individuo ms devil de la generacin anterior */
               index2= 0;
               minQuality.value= 0;
               minQuality.index= 0;
               for (index2; index2<5; index2++)
               { 
                  printf("%u %u %u %u\n\r", minQuality.index, minQuality.value, index2, selected[index2][1]);
                  if (selected[index2][1]>=minQuality.value)
                  {
                     minQuality.value= selected[index2][1];
                     minQuality.index= index2;
                  }
               }
               
               if (minQuality.value>=read_eeprom((index1*4)+3))
               {
                  selected[minQuality.index][0]= index1;
                  selected[minQuality.index][1]= read_eeprom((index1*4)+3);
               }
            }
  
         state= REPRODUCTION;
         break;      
     
      case REPRODUCTION:

         /* Cruzamos a los 5 individuos de la poblacin */
         chromosomeReproduction(selected[0][0], selected[1][0], populationIndex+0); 
         chromosomeReproduction(selected[0][0], selected[2][0], populationIndex+1); 
         chromosomeReproduction(selected[0][0], selected[3][0], populationIndex+2);
         chromosomeReproduction(selected[0][0], selected[4][0], populationIndex+3);
         chromosomeReproduction(selected[1][0], selected[2][0], populationIndex+4);
         chromosomeReproduction(selected[1][0], selected[3][0], populationIndex+5); 
         chromosomeReproduction(selected[1][0], selected[4][0], populationIndex+6); 
         chromosomeReproduction(selected[2][0], selected[3][0], populationIndex+7);
         chromosomeReproduction(selected[2][0], selected[4][0], populationIndex+8);
         chromosomeReproduction(selected[3][0], selected[4][0], populationIndex+9);
         
         populationLimit = populationLimit + 10;
         
         state= EVALUATION;
         break; 
   }
   
   if (populationIndex==55) /* Fin del experimento evolutivo*/
   { 
      motor1(0);
      motor2(0);
      bit_clear(puerto_b,0);
      bit_set(puerto_b,1); 
   }
   else /* La evolucin continua */
   { 
   /* Frecuencia de muestreo del bucle PID de 125Hz */
   set_timer1(64536);
   }
}

/******************************************************************************/
/*                                                                            */
/* experimentRestart                                                          */
/* Descripcin:      Lectura de la memoria EEPROM para mostrar el resultado   */
/*                   del experimento anterior. Pasados 10 segundos, preparamos*/
/*                   uno nuevo                                                */
/* Entrada:                                                                   */
/* Salida:                                                                    */
/* Auto:             Pedro Amador Daz                                        */
/* Fecha:            30/12/2011                                               */
/*                                                                            */
/******************************************************************************/

void experimentRestart()
{
   long index,seconds;
   
   /* Lectura EEPROM */ 
   for (index= 0; index<256; index=index+4)
   { 
      printf("Cromosoma %ld= %u, %u, %u, %u \n\r "(index/4)+1, read_eeprom(index), \
      read_eeprom(index+1), read_eeprom(index+2), read_eeprom(index+3));
   }
   
   /* Espera 10 segundos */  
   for (seconds=0; seconds<11; seconds++)
   {
      delay_ms(1000);
   }
   
   /* Borrado EEPROM */ 
   for (index= 0; index<256; index++)
   { 
      write_eeprom(index,0);
   }
   
   /* Generacin inicial de 5 individuos*/
   for (index= 0; index<5; index++)
   { 
      mutation();

      write_eeprom((index*4)+0,chromosome[0]);
      write_eeprom((index*4)+1,chromosome[1]);
      write_eeprom((index*4)+2,chromosome[2]);
      write_eeprom((index*4)+3,0);
      
      ++populationLimit;
   }
   
   populationIndex= populationLimit;
   state = REPRODUCTION;
      
}


/******************************************************************************/
/*                                                                            */
/* scannerReader                                                              */
/* Descripcin:      Lectura de los tres sensores infrarrojos                 */
/* Entrada:          Puntero al array que almacenar los valores ledos       */
/*                   por los sensores infrarrojos                             */
/* Salida:                                                                    */
/* Auto:             Pedro Amador Daz                                        */
/* Fecha:            24/10/2011                                               */
/*                                                                            */
/******************************************************************************/

void scannerReader(int *arrayScanner)
{
   int sensor;
   
   for (sensor=0; sensor<3; sensor++)
   {
      set_adc_channel(sensor); /* Cambio de canal A/D */
      delay_us(10); /* Espera tras cambiar de canal */
      arrayScanner[sensor]=255-read_adc();  /* Lectura valor analgico */
   }
}


/******************************************************************************/
/*                                                                            */
/* sensorsCalibration                                                         */
/* Descripcin:      Calibrado de las lecturas de los sensores infrarrojos    */
/* Entrada:          Puntero al array que almacena los valores ledos         */
/*                   por los sensores infrarrojos                             */
/* Salida:                                                                    */
/* Auto:             Pedro Amador Daz                                        */
/* Fecha:            24/10/2011                                               */
/*                                                                            */
/******************************************************************************/

void sensorsCalibration(int *arrayScanner)
{
   int sensor;
   
   for (sensor=0; sensor<3; sensor++)
   {
      /* Lectura inferior al mnimo posible */
      if (arrayScanner[sensor]<maxMinScanner[sensor][0])
         arrayScanner[sensor]=maxMinScanner[sensor][0];
      
      /* Lectura superior al mximo posible */
      if (arrayScanner[sensor]>maxMinScanner[sensor][1]) 
         arrayScanner[sensor]=maxMinScanner[sensor][1];
         
      /* Normalizado de las lecturas de los sensores infrarrojos */
      arrayScanner[sensor]=((float)(arrayScanner[sensor]-maxMinScanner[sensor][0])/
      (maxMinScanner[sensor][1]-maxMinScanner[sensor][0]))*255;
   }
}


/******************************************************************************/
/*                                                                            */
/* linePosition                                                               */
/* Descripcin:      Calculo de la posicin en la que se encuentra el robot   */
/*                   respecto a la lnea                                      */
/* Entrada:          Puntero al array que almacena los valores ledos         */
/*                   por los sensores infrarrojos                             */
/* Salida:           0 al 255, siendo 128 centrado                            */
/* Auto:             Pedro Amador Daz                                        */
/* Fecha:            24/10/2011                                               */
/*                                                                            */
/******************************************************************************/

int linePosition(int *arrayScanner)
{
  static int position=128;
  boolean calculate=false;
  int sensor;
  
  for (sensor=0; sensor<3; sensor++)
  {
      /* Solo se calcula si la lectura de la lnea tiene suficiente contraste */
      if (arrayScanner[sensor]>=127)
         calculate=true;
  }
  
  if (calculate==true)
      /* Calculo ponderado */
     position=(((0*(int32)arrayScanner[0])+(255*(int32)arrayScanner[1])+(510*(int32)arrayScanner[2]))/
           ((int32)arrayScanner[0]+(int32)arrayScanner[1]+(int32)arrayScanner[2]))/2;
           
 return position;
}

/******************************************************************************/
/*                                                                            */
/* reactiveControlPID                                                         */
/* Descripcin:      Inteligencia reactiva modulada mediante control PID      */
/* Entrada:          Posicin del robot respecto a la lnea. Del 0 al 255,    */
/*                   siendo 128 centrado                                      */
/* Salida:                                                                    */
/* Auto:             Pedro Amador Daz                                        */
/* Fecha:            30/12/2011                                               */
/*                                                                            */
/******************************************************************************/

#INT_TIMER1
void reactiveControlPID()
{  
   signed long PID= 0, errorDerivative= 0;
   static signed long errorSum, errorLast;
    
   /* Calculo de la componente integral */
   errorSum = errorSum + error;
   
   /* Tratamiento del desbordamiento de la variable errorSum */
   if (errorSum>=32000)
      errorSum = 32000;
   else if (errorSum<=-32000)
      errorSum = -32000;
      
   /* Clculo de la componente derivada */   
   errorDerivative = error - errorLast;
   errorLast = error;
   
   
   PID = (signed long)((Kp*error)+(Ki*(errorSum/125))+(Kd*(errorDerivative*125)));
   
   /* Traccin con direccin diferencial */
   motor1(PWM+PID);
   motor2(PWM-PID);
   
   /* Actualizacin de la mquina de estados que conforma el algoritmo gentico */
   geneticAlgorithm();
}

/******************************************************************************/
/*                                                                            */
/* reactiveControl                                                            */
/* Descripcin:      Inteligencia reactiva simple                             */
/* Entrada:          Posicin del robot respecto a la lnea. Del 0 al 255,    */
/*                   siendo 128 centrado                                      */
/* Salida:                                                                    */
/* Auto:             Pedro Amador Daz                                        */
/* Fecha:            24/10/2011                                               */
/*                                                                            */
/******************************************************************************/

void reactiveControl (int position)
{
   if (position < 85) /* Gira a la derecha */
   {
      /* Led de color verde */
      bit_set(puerto_b,0);
      bit_clear(puerto_b,1);
         
      /* Motor derecho frenado */
      motor1(-10);
      motor2(75);
   }
   else if (position < 170) /* Sigue recto */
   {
      /* Led de color amarillo */
      bit_set(puerto_b,0);
      bit_set(puerto_b,1);
         
      /* PWM motores izquierdo y derecho a la misma velocidad */
      motor1(75);
      motor2(75);
      }
   else /* Gira a la izquierda */
   {
      /* Led de color rojo */
      bit_clear(puerto_b,0);
      bit_set(puerto_b,1);
         
      /* Motor izquierdo frenado */
      motor1(75);
      motor2(-10);
   }
}


/******************************************************************************/
/*                                                                            */
/* main                                                                       */
/* Descripcin:      Funcin que inicializa el sistema y contiene el bucle    */
/*                   principal (motor del sistema reactivo sin control PID)   */
/* Entrada:                                                                   */
/* Salida:                                                                    */
/* Auto:             Pedro Amador Daz                                        */
/* Fecha:            24/10/2011                                               */
/*                                                                            */
/******************************************************************************/

void main()
{   
   /* Leemos el resultado del experimento anterior 
      y pasados 10 segundos iniciamos uno nuevo */
   experimentRestart();
   
   /* Inicializacin de los timers */
   setup_timer_1(T1_INTERNAL | T1_DIV_BY_8);
   enable_interrupts(INT_TIMER1);
   enable_interrupts(GLOBAL);
   
   set_tris_a(0xFF); /* Puerto A como entreada */
   set_tris_b(0x00); /* Puerto B como salida */
   set_tris_c(0x00); /* Puerto C como salida */
   
   /* Puertos A0 A1 A2 A3 A4 entradas analgicas */
   setup_adc( ADC_CLOCK_DIV_8 );
   setup_adc_ports( ALL_ANALOG );
   
   /* Inicializacin de los PWM */
   setup_timer_2(T2_DIV_BY_1,255,1);
   setup_ccp1(ccp_pwm);
   setup_ccp2(ccp_pwm);
  
   /* Inicializacin del display */
   bit_set(puerto_b,0);
   bit_clear(puerto_b,1);
   
      set_timer1(64536);   set_timer1(64536);set_timer1(64536);
      
   while (true)
   {
      /* Lectura de los valores recogidos por los sensores de infrarrojos */
      scannerReader(arrayScanner);
   
      /* Calibrado de los valores recogido por los sensores */
      sensorsCalibration(arrayScanner);
      
      /* Clculo del error */   
      error = (signed int)linePosition(arrayScanner)-128;
      
      /* Para un control reactivo sin control PID, comentar las lineas de cdigo
         referentes al timer1, calculo de error y descomentar las siguientes lineas*/
      //reactiveControl (linePosition(arrayScanner))
   }
}
